@@ -1,10 +1,7 @@
 from .aggregations import Aggregate, GroupBy, Reduce, ReduceFuncs
 from .base import (
     And,
-    BaseCollectionConversion,
     BaseConversion,
-    CachingConversion,
-    Call,
     CallFunc,
     ConversionException,
     ConverterOptionsCtx,
@@ -30,6 +27,7 @@ from .base import (
     TupleComp,
     ensure_conversion,
 )
+from .joins import _JoinConditions, join
 
 
 __all__ = ["conversion"]
@@ -74,6 +72,10 @@ class _Conversion:
     reduce = Reduce
     aggregate = staticmethod(Aggregate)
 
+    join = staticmethod(join)
+    LEFT = _JoinConditions.LEFT
+    RIGHT = _JoinConditions.RIGHT
+
     def __call__(self, obj: object):
         """Shortcut for ``ensure_conversion``"""
         return ensure_conversion(obj)
